"""
负责资源导入、模块检查
"""
import os
import cv2

from stagesep2.logger import logger
from stagesep2.utils import *


def path_to_name(file_path):
    """ full path -> file name """
    _, name = os.path.split(file_path)
    name, _ = os.path.splitext(name)
    return name


def is_path_existed(file_path):
    """ check if file is existed """
    return os.path.isfile(file_path)


def frame_prepare(frame):
    gray_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    blur_gray_frame = cv2.medianBlur(gray_frame, 3)
    return blur_gray_frame


class TemplatePicture(object):
    def __init__(self, pic_path):
        if not is_path_existed(pic_path):
            raise FileNotFoundError('file not existed: {}'.format(pic_path))

        self.pic_name = path_to_name(pic_path)
        self.pic_path = pic_path
        self.cv_object = frame_prepare(cv2.imread(self.pic_path))


class TemplateManager(object):
    TAG = 'TemplateManager'
    # match template 需要模板图片
    # 该视频需要的模板图片会被放置在此处
    _match_template_pic_dict = dict()
    # eg:
    # { pic_name: TemplatePicture(pic_path), }

    def add(self, pic_path):
        new_pic = TemplatePicture(pic_path)
        new_pic_name = new_pic.pic_name
        self._match_template_pic_dict[new_pic_name] = new_pic
        logger.info(self.TAG, msg='LOAD PICTURE', path=pic_path, name=new_pic_name)

    def remove(self, pic_name):
        if pic_name in self._match_template_pic_dict:
            del self._match_template_pic_dict[pic_name]
            return True
        logger.warn(self.TAG, msg='no pic named {}'.format(pic_name))
        return False

    def get_dict(self):
        return self._match_template_pic_dict


class SSVideo(object):
    """ video object """

    def __init__(self, video_path):
        if not is_path_existed(video_path):
            raise FileNotFoundError('file not existed: {}'.format(video_path))

        self.video_name = path_to_name(video_path)
        self.video_path = video_path

        # add template example:
        # ssv = SSVideo('some_path/123.mp4')
        # ssv.template_manager.add('some_path/123.png')
        self.template_manager = TemplateManager()

        # degree = rotate * 90, 逆时针
        self._rotate = 0

        # video info
        # total frame count
        self.total_frame = None

        # first and last frame
        self.first_frame = None
        self.last_frame = None

    @property
    def rotate(self):
        return self._rotate

    @rotate.setter
    def rotate(self, value):
        if not isinstance(value, int):
            raise TypeError('rotate should be int')
        self._rotate = value
        self.load_video_info()

    def load_video_info(self):
        # TODO need more info?
        # get info from video
        with video_capture(self) as video_src:
            total_frame = video_src.get(cv2.CAP_PROP_FRAME_COUNT)

            _, first_frame = video_src.read()
            video_src.set(1, total_frame - 1)
            _, last_frame = video_src.read()

            # prepare, and rotate
            first_frame, last_frame = [
                frame_prepare(rotate_pic(each, self._rotate))
                for each in (first_frame, last_frame)]

        # init
        self.first_frame = first_frame
        self.last_frame = last_frame
        self.total_frame = total_frame


class VideoManager(object):
    """
    Analyser需要的信息都应该在此处被导入
    例如 作为分析主体的 视频
    例如 match template需要的 模板图片
    """
    TAG = 'VideoManager'

    # 待测视频会被添加到这里
    # 在分析开始时，会遍历此字典
    video_dict = dict()
    # eg:
    # { video_name: SSVideo(video_path), }

    def __init__(self):
        raise NotImplementedError('should not init')

    @classmethod
    def add(cls, video_path):
        new_video = SSVideo(video_path)
        new_video_name = new_video.video_name
        cls.video_dict[new_video_name] = new_video
        logger.info(cls.TAG, msg='LOAD VIDEO', path=video_path, name=new_video_name)
        return new_video

    @classmethod
    def remove(cls, video_name):
        if video_name in cls.video_dict:
            del cls.video_dict[video_name]
            return True
        logger.warn(cls.TAG, msg='no video named {}'.format(video_name))
        return False



from skimage.measure import compare_ssim

from stagesep2.analyser.base import BaseAnalyser


class TrendAnalyser(BaseAnalyser):
    """ trend analyser """
    name = 'trend'

    @classmethod
    def run(cls, frame, ssv):
        """
        return a dict, eg: {'first': 0.9, 'last': 0.2}

        :param frame:
        :param ssv:
        :return:
        """
        first_sim = compare_ssim(ssv.first_frame, frame)
        last_sim = compare_ssim(ssv.last_frame, frame)
        result_dict = {
            'first': first_sim,
            'last': last_sim,
        }
        return result_dict



"""
处理分析结果

应至少支持：
- 文件（json）
- 对象
- 数据库交互
"""
import uuid
import json
import os

from stagesep2.logger import logger


class ResultRow(object):
    def __init__(
        self,
        # must
        result_id,
        video_name,
        frame_id,
        current_time,
    ):
        # task id
        self.result_id = result_id
        # video name (id), a task may contains more than one video
        self.video_name = video_name
        # current frame number
        # 当前的结果对应的帧编号
        self.frame_id = frame_id
        # current time
        # 当前的结果对应的帧，在视频中对应的时间
        self.current_time = current_time

    def add_analyser_result(self, name, result):
        """
        add result of analyser

        :param name: analyser name
        :param result: result of analyser
        :return:
        """
        # analyse result
        # ocr: list, eg: ['some_word', 'other_word']
        # match_template: dict, eg: {'pic1': 0.85, 'pic2': 0.90}

        self.__dict__[name] = result

    def __str__(self):
        return json.dumps(self.__dict__)

    __repr__ = __str__


class ResultReporter(object):
    TAG = 'ResultReporter'

    def __init__(self):
        self.result_id = str(uuid.uuid1())
        self._row_list = list()

    def add_row(self, new_row):
        self._row_list.append(new_row)

    def export(self, file_path):
        """ export result to json file """
        if os.path.isfile(file_path):
            logger.warn(self.TAG, msg='File "{}" already existed'.format(file_path))
            file_path = os.path.join(os.path.dirname(file_path), self.result_id + '.json')
        elif os.path.isdir(file_path):
            logger.warn(self.TAG, msg='Path "{}" is a directory'.format(file_path))
            file_path = os.path.join(file_path, self.result_id + '.json')

        with open(file_path, 'w+') as json_file:
            json_file.write(str(self.data))
            logger.info(self.TAG, msg='Result saved in "{}"'.format(file_path))

    @property
    def data(self):
        """ return data, consisted by pyobject """
        return self._row_list



import cv2

from stagesep2.analyser.base import BaseAnalyser
from stagesep2.config import MatchTemplateConfig


class MatchTemplateAnalyser(BaseAnalyser):
    """ match-template analyser """
    name = 'match_template'

    @classmethod
    def run(cls, frame, ssv):
        result_dict = dict()
        template_pic_dict = ssv.template_manager.get_dict()
        for each_pic_name, each_pic in template_pic_dict.items():
            # TODO match_template not always very match
            res = cv2.matchTemplate(frame, each_pic.cv_object, MatchTemplateConfig.cv_method)
            min_val, max_val, _, _ = cv2.minMaxLoc(res)
            result_dict[each_pic_name] = {
                'min': min_val,
                'max': max_val,
            }
        return result_dict



"""
analyser

opencv frame -> result
"""
from stagesep2.analyser.ocr import OCRAnalyser
from stagesep2.analyser.match_template import MatchTemplateAnalyser
from stagesep2.analyser.trend import TrendAnalyser


ANALYSER_DICT = {
    'ocr': OCRAnalyser,
    'match_template': MatchTemplateAnalyser,
    'trend': TrendAnalyser,
}



import subprocess
import tempfile
import os
import cv2
import uuid
import jieba
import platform

from stagesep2.analyser.base import BaseAnalyser
from stagesep2.config import OCRConfig, NormalConfig


def content_filter(old_content):
    """ remove unused content and rebuild a word list """
    new_content = old_content.replace(' ', '').replace('\n', '').replace('\r', '')
    return list(jieba.cut(new_content))


class OCRAnalyser(BaseAnalyser):
    """ ocr analyser """
    # TODO 大量的io开销
    name = 'ocr'

    @classmethod
    def exec_tesseract(cls, src, dst):
        cmd = ['tesseract', src, dst, '-l', OCRConfig.lang]
        need_shell = platform.system() == 'Windows'
        tesseract_process = subprocess.Popen(
            cmd,
            shell=need_shell,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        tesseract_process.communicate(timeout=5)
        if tesseract_process.returncode:
            error_msg = tesseract_process.stderr.read()
            raise RuntimeError('tesseract error: {}'.format(error_msg))

    @classmethod
    def run(cls, frame, *args, **kwargs):
        """
        run ocr analyser

        1. write frame to file
        2. execute tesseract to analyse it
        3. and get its result
        4. delete temp file and return result

        :param frame:
        :return:
        """
        # create temp picture file
        temp_pic = tempfile.NamedTemporaryFile(delete=False, suffix='.png')
        temp_pic_path = temp_pic.name
        temp_pic.close()
        # tesseract will auto create result file
        # and add '.txt' after its name!
        temp_result_path = os.path.join(NormalConfig.PROJECT_PATH, str(uuid.uuid1()))
        real_temp_result_path = temp_result_path + '.txt'

        # write in
        cv2.imwrite(temp_pic_path, frame)
        # execute tesseract
        cls.exec_tesseract(temp_pic_path, temp_result_path)
        # get result
        with open(real_temp_result_path, encoding='utf-8') as result_file:
            result = result_file.read()
        # content filter
        result = content_filter(result)
        # remove temp files
        os.remove(temp_pic_path)
        os.remove(real_temp_result_path)
        return result



from setuptools import setup, find_packages


setup(
    name='stagesep2',
    version='0.1.2',
    description='detect stages in video',
    author='williamfzc',
    author_email='fengzc@vip.qq.com',
    url='https://github.com/williamfzc/stagesep2',
    packages=find_packages(),
    install_requires=[
        'opencv-python',
        'structlog',
        'numpy',
        'jieba',
        'scikit-image'
    ]
)


"""
public functions
"""
import numpy as np
import cv2
import contextlib

from stagesep2.analyser import ANALYSER_DICT


def check_analyser(analyser_list):
    """ check if analyser existed, and return list of runnable analyser """
    new_analyser_list = list()
    for each in analyser_list:
        if each not in ANALYSER_DICT:
            raise NotImplementedError('analyser {} not found'.format(each))
        new_analyser_list.append(ANALYSER_DICT[each])
    return new_analyser_list


def rotate_pic(old_pic, rotate_time):
    """ 帧逆时针旋转 90*rotate_time 度 """
    new_pic = np.rot90(old_pic, rotate_time)
    return new_pic


@contextlib.contextmanager
def video_capture(ssv):
    """ 打开视频的上下文控制 """
    video_cap = cv2.VideoCapture(ssv.video_path)
    yield video_cap
    video_cap.release()


__all__ = [
    'check_analyser',
    'rotate_pic',
    'video_capture',
]
